ad_library {
    A library of functions to generate forms for acs_objects from stored 
    metadata.
    
    The API manipulates two concepts - forms and elements.  Forms are 
    mapped to object_types.  Each object_type can have several named forms 
    mapped to it.  Each form contains several elements which correspond to 
    the attributes of its object_type.  

    It's possible, for example, to create two forms - 'admin' and 'public'.  
    The public form could contain elements for the attributes that may be 
    modified by public users whereas the admin form could contain 
    additional elements that admin users can edit.

    Elements have associated parameters which control how the html 
    representation is displayed, how values and options are retrieved and
    how submitted values are validated.  
}

namespace eval dtype {}
namespace eval dtype::form {}
namespace eval dtype::element {}
namespace eval dtype::layout {}
namespace eval dtype::spec {}

# TODO  Add caching without util_memoize when there is time and can figure
# TODO  out the dependencies

#
# Forms
#

ad_proc -public dtype::form::create {
    {-layout {}}
    {-widget {}}
    {-custom {}}
} {
    Creates a dynamic form.
} {
    return [db_exec_plsql create_form {}]
}

ad_proc -public dtype::form::delete {
    {-form_id:required}
} {
    Deletes a dynamic form.
} {
    db_exec_plsql delete_form {}
}

ad_proc -public dtype::form::add_element {
    {-form_id:required}
    {-element:required}
    {-widget:required}
    {-layout {}}
    {-custom {}}
} {
    Create a top level section in a form.
} {
    set parent_id [dtype::form::get_root_element_id -form_id $form_id]

    return [dtype::element::add_element -element $element \
        -parent_id $parent_id \
        -widget $widget \
        -layout $layout \
        -custom $custom]
}

ad_proc -public dtype::form::elements {
    {-form_id:required}
    {-local:boolean}
    {-ulevel 1}
    {-multirow:required}
} {
    Returns the depth first span of the metadata for the specified form as a 
    multirow.

    @param form_id specifies the form to get
    @param multirow the name of the multirow to populate
} {
    set keys [list \
        element_id \
        element \
        widget \
        custom \
        parent_id \
        level \
        layout_id \
        filename \
        resource_dir \
        attribute_id \
        attribute_name \
        attribute_pretty_name \
        attribute_object_type \
        attribute_datatype \
        attribute_default_value \
    ]

    set multirow_cmd "template::multirow"

    if {$local_p} {
        append multirow_cmd " -local -ulevel [expr {$ulevel + 1}]"
    }

    eval "$multirow_cmd create \$multirow $keys"

    set metadata [dtype::form::elements_list -form_id $form_id]

    foreach element $metadata {
        eval "$multirow_cmd append \$multirow $element"
    }
}

ad_proc -private dtype::form::elements_list {
    {-form_id:required}
} {
    Returns the depth first span of the metadata for the specified object_type 
    and form as a multirow.

    @param object_type the object type whose metadata will define the form
    @param form specifies the stored object form to use
} {
    return [db_list_of_lists select_form_metadata {}]
}

ad_proc -private dtype::form::get_root_element_id {
    {-form_id:required}
} {
    Get the id of the root element for a form.
} {
    return [db_string get_element_id {}]
}

ad_proc -public dtype::form::map_type {
    {-form_id:required}
    {-form:required}
    {-object_type:required}
} {
    Maps an object type to a form.
} {
    db_exec_plsql map_type {}
}

ad_proc -public dtype::form::unmap_type {
    {-form_id:required}
    {-object_type:required}
} {
    Unmaps an object type from a form.
} {
    db_dml unmap_type {}
}

ad_proc -public dtype::form::element_params {
    {-form_id:required}
    {-local:boolean}
    {-ulevel 1}
    {-multirow:required}
} {
    Create a multirow of the element parameters for the specified form indexed
    by element_id.

    @param object_type the object type whose metadata will define the form
    @param form specifies the stored object form to use
    @param multirow the name of the multirow to populate
} {
    set keys [list \
      value_id \
      element_id \
      param_id \
      name \
      param_type \
      param_source \
      is_required \
      default_value \
      value \
    ]

    set multirow_cmd "template::multirow"

    if {$local_p} {
        append multirow_cmd " -local -ulevel [expr {$ulevel + 1}]"
    }

    eval "$multirow_cmd -key element_id create \$multirow $keys"

    set metadata [dtype::form::element_params_list -form_id $form_id]

    foreach param $metadata {
        eval "$multirow_cmd append \$multirow $param"
    }
}

ad_proc -private dtype::form::element_params_list {
    {-form_id:required}
} {
    Returns a list of lists with the widget parameter metadata for the 
    specified object_type and form.

    @param form specifies the stored object form to use
} {
    return [db_list_of_lists select_form_metadata {}]
}

ad_proc -private dtype::form::flush_cache {
    {-form {}}
} {
    Flushes the cache of dtype::form calls for a given object type.
} {
}

#
# Elements
#

ad_proc -public dtype::element::add_element {
    {-element:required}
    {-parent_id:required}
    {-widget {}}
    {-layout {}}
    {-custom {}}
} {
    Creates an element for on specified form.
} {
    set element_id [db_exec_plsql create_element {}]
}

ad_proc -public dtype::element::get {
    {-element_id:required}
    {-array:required}
} {
    Gets the data about an element.
} {
    upvar $array result

    db_1row select_element_metadata {} -column_array result
}

ad_proc -public dtype::element::delete {
    {-element_id:required}
} {
    Deletes an element.
} {
    db_exec_plsql delete_element {}
}

ad_proc -public dtype::element::add_parameter {
    {-element_id:required}
    {-param:required}
    {-value:required}
    {-source literal}
} {
    Adds a parameter to the template::element::create proc for a dtype form
    element.
} {
    db_exec_plsql add_parameter {}
}

ad_proc -public dtype::element::create_or_replace_parameter {
    {-element_id:required}
    {-param:required}
    {-value:required}
    {-source literal}
} {

    The code-generation scripts need to be able to add a parameter without
    worrying about if it's already been created. Why? When we create an
    element, our code checks to see if there are any 'required' parameters
    for that element. If so, it automatically adds those parameters with
    their default values. Later, when we dynamically generate code to
    recreate elements (code-generation), it's easier just to write code
    which 'create_or_replaces' parameters, rather then check each parameter
    to see if it's a required parameter and whether the current value is
    equal to the default or not.

    @author Vinod Kurup
} {
    if { [db_0or1row param_exists {}] } {
        # update param
        db_dml update_param {}
    } else {
        # add param
        dtype::element::add_parameter \
            -element_id $element_id \
            -param $param \
            -value $value \
            -source $source
    }
}

ad_proc -public dtype::element::delete_parameter {
    {-element_id:required}
    {-param:required}
} {
    Deletes a parameter from the template::element::create proc for a dtype 
    form element.
} {
    db_exec_plsql delete_parameter {}
}

ad_proc -public dtype::element::map_attribute {
    {-element_id:required}
    {-attribute_id:required}
} {
    Maps an attribute to a form element.
} {
    db_exec_plsql map_attribute {}
}

ad_proc -public dtype::element::unmap_attribute {
    {-element_id:required}
    {-attribute_id:required}
} {
    Unmaps an attribute from a form element.
} {
    db_dml unmap_attribute {}
}

#
# Layouts
#

ad_proc -public dtype::layout::create {
    {-layout:required}
    {-name:required}
    {-description:required}
    {-filename:required}
    {-resource_dir:required}
    {-type form}
} {
    Create a top level section in a form.
} {
    return [db_exec_plsql create_layout {}]
}

ad_proc -public dtype::layout::delete {
    {-layout_id:required}
} {
    Deletes a section.
} {
    db_exec_plsql delete_layout {}
}

ad_proc -public dtype::layout::get_layout_id {
    {-layout:required}
} {
    Gets the id of a layout.
} {
    return [db_string get_layout_id {}]
}



#
# Form generation and processing procs
#

ad_proc -private dtype::form::get_id {
    {-form:required}
    {-object_type:required}
} {
    Gets the id of a form from the name and associated object_type.
} {
    return [db_string get_form_id {}]
}

ad_proc -private dtype::form::get {
    {-form_id:required}
} {
    Return an array list of the form details for a form_id.
} {
    db_1row get_form {} -column_array form

    return [array get form]
}

ad_proc -public dtype::form::generate {
    {-dtype_form:required}
    {-object_type ""}
    {-object_id ""}
    {-ats_form:required}
    {-prefix ""}
    {-overrides {}}
} {
    Adds the elements of the specified dynamic form to the specified template 
    form.

    This function is used for both add and edit forms.  It determines which is
    appropriate based on whether an object_id or an object_type is supplied.  
    Only one of object_type or object_id may be supplied.

    @param object_id the object represented in the form.  If set the form is
           assumed to be an edit form, otherwise it is assumed to be an object
           create form
    @param object_type the object type whose metadata will define the form
    @param dtype_form specifies the dynamic form to use
    @param ats_form the name of the template::form to add the elements to (will 
           create the form if it doesn't already exist)
    @param prefix prefix for each attribute name to avoid collisions
    @param overrides key value pairs that will override the initial value of
           a form element irrespective of the default value in the form 
           metadata of the value in the object being edited
} {
    if {![template::form exists $ats_form]} {
        template::form create $ats_form
    }

    ############################################################
    # Determine the form action
    #
    array set override $overrides

    if {[string equal $object_id ""]} { 
        set context(action) create

        if {![info exists override(object_id)]} {
            set override(object_id) [db_nextval acs_object_id_seq]
        }

        set context(object_id) $override(object_id)
    } else {
        set context(action) edit
        set object_type [acs_object_type $object_id]

        if {[info exists override(object_id)]} {
            error "Cannot override object_id for an existing object"
        }

        set context(object_id) $object_id
    }

    set context(user_id) [ad_conn user_id]

    set form_id [dtype::form::get_id \
        -form $dtype_form \
        -object_type $object_type]

    # Record the id of the dtypes form
    template::element create $ats_form $ats_form:dtype:${prefix}id \
        -widget hidden \
        -datatype text \
        -sign \
        -value $form_id  

    # Record the action in the form
    template::element create $ats_form $ats_form:dtype:${prefix}action \
        -widget hidden \
        -datatype text \
        -sign \
        -value $context(action)  

    # Record the object_id in the form
    template::element create $ats_form $ats_form:dtype:${prefix}object_id \
        -widget hidden \
        -datatype integer \
        -sign \
        -value $context(object_id)  

    if {[string equal $context(action) edit]} {
        # Get existing object data
        dtype::get_object -object_id $object_id -array object
    }

    ############################################################
    # Generate the template::form elements
    #
    dtype::form::elements -local -form_id $form_id -multirow elements
    dtype::form::element_params -local -form_id $form_id -multirow params

    set previous_level 0
    set nested_name [list]

    # element values generated by custom callback and passed down to element
    array set custom_values [list]

    # uncomment for debug usefulness
    #template::multirow -local log elements

    set size [template::multirow -local size elements]
    for {set e 1} {$e <= $size} {incr e} {
        template::multirow -local get elements $e

        if {$elements(level) <= $previous_level} {
            # Replace the previous nested name elements with the current 
            # element name
            set nested_name [lreplace $nested_name \
                [expr {[llength $nested_name] - ($previous_level - $elements(level)) - 1}] \
                end \
                $elements(element)]
        } else {
            lappend nested_name $elements(element)
        }

        # custom callback to fill in the form values
        # pass name of array to add data to, since there may be multiple items
        if {![string equal $elements(custom) ""] 
            && [callback::impl_exists \
                -callback dtype::form::get_db_values \
                -impl $elements(custom)]} {

            # the object will contain the id for the custom element
            # check it exists before calling the callback
            if {![string equal $elements(attribute_name) ""]} {
                if {[info exists object($elements(attribute_name))]} {
                    callback -impl $elements(custom) \
                        dtype::form::get_db_values \
                            -attribute_name $elements(attribute_name) \
                            -object_value $object($elements(attribute_name)) \
                            -values custom_values
                }
            } 
        }

        if {![string equal $elements(widget) ""]} {
            # Add an element to the template form

            if {[info exists override($elements(element))]} {
                set elm_overrides $override($elements(element))
            } else {
                set elm_overrides [list]
            }

            set defaults [list]
            if {![string equal $elements(attribute_name) ""]} {
                lappend defaults label $elements(attribute_pretty_name)

                if {[info exists object($elements(attribute_name))]} {
                    lappend defaults value $object($elements(attribute_name))
                }
            } elseif {[info exists custom_values($elements(element))]} {
                # no attribute name, but a value in the custom values
                lappend defaults value $custom_values($elements(element))
            }

            if {![string equal $elements(attribute_datatype) ""]} {
                lappend defaults datatype $elements(attribute_datatype)
            }

            dtype::element::generate -local \
                -element elements \
                -params params \
                -ats_form $ats_form \
                -prefix $prefix \
                -context context \
                -overrides $elm_overrides \
                -defaults $defaults \
                -custom_args [list -styles $nested_name]
        }
    }
}

ad_proc -private dtype::element::generate {
    {-element:required}
    {-params:required}
    {-ats_form:required}
    {-local:boolean}
    {-ulevel 1}
    {-prefix ""}
    {-context:required}
    {-overrides ""}
    {-defaults ""}
    {-custom_args {}}
} {
    Adds an element of the specified object form to the specified template form.

    @param element the element to be generated
    @param params the params for the element being generated
    @param ats_form the name of the template::form to add the elements to
    @param prefix optional prefix for each element name to avoid collisions
    @param value an explicit initial value for the element
} {
    upvar $element l_element
    upvar $context l_context

    template::multirow -local=$local_p \
        -ulevel $ulevel \
        upvar \
        $params \
        l_param_values

    unset params

    array set l_overrides $overrides
    array set l_defaults $defaults
    
    # TODO: this should be cached - it's getting run a lot
    dtype::proc::params -proc "template::element::create" \
        -multirow params \
        -local 
    
    # Create the main element create line
    set element_create_cmd "template::element create \
          \$ats_form \
          \${prefix}\$l_element(element) \
          -widget \$l_element(widget)"

    # Build an index of the parameter values
    set param_values [template::multirow -local \
          -key $l_element(element_id) \
          rows \
          l_param_values]

    foreach param_value $param_values {
        template::multirow -local \
            get \
            l_param_values \
            $param_value

        set value_idx($l_param_values(name)) $param_value
    }

    # Create an array for spec switches, which are built up from several
    # parameters
    array set spec_switches [list]
    set switches [list]
    set bool_switches [list]

    array set command_values [list]

    # Generate the parameters
    set size [template::multirow -local size params]
    for {set p 1} {$p <= $size} {incr p} {
        template::multirow -local get params $p

        if {[info exists l_overrides($params(name))]} {

            # Use the passed in override value for the current parameter
            set value $l_overrides($params(name))

        } elseif {[info exists value_idx($params(name))]} {

            # Use the parameter value stored in the database
            template::multirow -local \
                get \
                l_param_values \
                $value_idx($params(name))

            set value [dtype::proc::param_value \
                -context l_context \
                -parameter params \
                -value l_param_values]

        } elseif {[info exists l_defaults($params(name))]} {

            # Use the passed in default value
            set value $l_defaults($params(name))

        } elseif {[info exists params(default_value)] 
                  && ![string equal $params(default_value) ""]} {

            # Use the default value stored in the database
            set value $params(default_value)

        }

        if {[info exists value]} {
            switch -- $params(param_type) {
                switch:onevalue {
                    switch -- $params(name) {
                        value {
                            switch -- $l_element(widget) {
                                date {
                                    set command_values(value) [template::util::date::from_ansi $value]
                                }
                                telecom_number {
                                    # convert the id into a meaningful value for the widget
                                    if { ![string equal $value {}] } {
                                        set command_values(value) [telecom_number::get -number_id $value]
                                    } else {
                                        set command_values(value) {}
                                    }
                                }
                                default {
                                    set command_values(value) $value
                                }
                            }

                            append switches [dtype::form::value_switch \
                                -widget $l_element(widget) \
                                -value "\$command_values(value)"]
                        }
                        default {
                            set command_values($params(name)) $value
                            append switches " -$params(name) \$command_values($params(name))"
                        }
                    }
                }
                switch:boolean {
                    if {[template::util::is_true $value]} {
                        lappend bool_switches "-$params(name)"
                    }
                }
                switch:spec {
                    set name_list [split $params(name) ","]
                    set major [lindex $name_list 0]
                    set minor [lindex $name_list 1]

                    lappend spec_switches($major) $minor $value
                }
            }

            unset value
        }
    }
        
    foreach name [array names spec_switches] {
        append element_create_cmd " -$name \$spec_switches($name)"
    }

    eval "$element_create_cmd $bool_switches $switches $custom_args"
}

ad_proc -public dtype::form::process {
    {-prefix ""}
    {-ats_form:required}
    {-defaults {}}
    args
} {
    Process a form containing a dtype form.
} {
    set form_id [template::element::get_value \
        $ats_form \
        $ats_form:dtype:${prefix}id]
    
    array set form [dtype::form::get -form_id $form_id]
    
    set action [template::element::get_value \
        $ats_form \
        $ats_form:dtype:${prefix}action]
    
    set object_id [template::element::get_value \
        $ats_form \
        $ats_form:dtype:${prefix}object_id]
    
    set supertypes "$form(object_type) [acs_object_type::supertypes -subtype $form(object_type)]"
    for {set i 0} {$i < [llength $supertypes]} {incr i} {
        set type [lindex $supertypes $i]

        if {[callback::impl_exists \
                -callback dtype::form::process \
                -impl $type]} {

            set results [callback -impl $type dtype::form::process \
                -form_id $form_id \
                -prefix $prefix \
                -ats_form $ats_form \
                -dtype_form $form(form) \
                -action $action \
                -object_type $form(object_type) \
                -object_id $object_id \
                -defaults $defaults \
                $args]

            return [lindex $results 0]

        }
    }

    return ""
}

ad_proc -callback dtype::form::process {
    {-prefix ""}
    {-ats_form:required}
    {-form_id:required}
    {-dtype_form:required}
    {-action:required}
    {-object_type:required}
    {-object_id:required}
    {-defaults {}}
    args
} {
    Process a form for a specific base object type.
} -

ad_proc -callback dtype::form::get_db_values {
    {-attribute_name:required}
    {-object_value:required}
    {-values:required}
} {
    Get values from database for custom widgets

    @param object_value the value stored in the dynamic object for this element
    @param values name of array to write values into
} -


ad_proc -callback dtype::form::process -impl acs_object {
} {
    Process a dynamic type form submission created by a function such as
    dtype::form::generate.  

    @param prefix the prefix for each element in the form
    @param ats_form the name of the template::form used
    @param form_id the form_id of the dtype form
    @param dtype_form specifies the stored object form used
    @param action whether the form is creating an object or editing one
    @param object_type the object type whose metadata will define the form
    @param object_id of the object being manipulated
    @param defaults default values to use for attributes
    
    @see dtype::form::generate
} {
    acs_object_type::get -object_type $object_type -array type_db

    array set default $defaults

    #######################################################
    # Setup default values for object creation
    #
    if {[string equal $action create]} {
        set default(object_type) $object_type

        if {![info exists default(creation_user)]} {
            set default(creation_user) [ad_conn user_id]
        }

        if {![info exists default(creation_ip)]} {
            set default(creation_ip) [ad_conn peeraddr]
        }
    }
    
    #######################################################
    # Setup the context for custom element callbacks
    #
    array set context [list \
        action $action \
        object_id $object_id \
        object_type $object_type \
        user_id [ad_conn user_id] \
    ]

    if {[string equal $action edit]} {
        # Get existing object data
        dtype::get_object -object_id $object_id -array object

        set context(old_values) [array get object]
    }

    #######################################################
    # Get the elements in the form
    #
    array set values [list]

    dtype::form::elements -local -form_id $form_id -multirow elements

    dtype::form::pre_process -prefix $prefix \
        -ats_form $ats_form \
        -elements elements \
        -local \
        -values values \
        -context context

    #######################################################
    # Perform the insert or update
    #
    template::multirow -local -key attribute_id index elements 

    # get the attribute metadata for the object type
    dtype::get_attributes -local -name $object_type \
        -start_with acs_object \
        attributes

    set column_names [list]
    set column_values [list]

    set size [template::multirow -local size attributes]
    for {set i 1} {$i <= $size} {incr i} {
        template::multirow -local get attributes $i
        
        # Get the elements row number for this attribute if it exists
        set element_row [lindex [template::multirow -local \
                    -key $attributes(attribute_id) \
                    rows \
                    elements] \
                0]

        if {![string equal $element_row ""]} {
            template::multirow -local get elements $element_row

            # This attribute is represented in the form so use the value given
            # in the form

            if {[info exists values($elements(element))]} {
                lappend column_names $attributes(name)
                set bind_$attributes(name) $values($elements(element))

                lappend column_values [dtype::sql_value \
                    $attributes(datatype) \
                    bind_$attributes(name)]
            }
        }

        if {![info exists bind_$attributes(name)]} {
            # There was no value generated by the form for this attribute so 
            # there are several ways to determine a default value
        
            if {[info exists default($attributes(name))]} {
        
                # A default was provided for this attribute so we use it
            
                lappend column_names $attributes(name)
                set bind_$attributes(name) $default($attributes(name))

                lappend column_values [dtype::sql_value \
                    $attributes(datatype) \
                    bind_$attributes(name)]

            } elseif {[string equal $action create]} {
                if {![string equal $attributes(default_value) ""]} {

                    # We are creating a new object so we use the default 
                    # value for the attribute supplied in the type metadata
                
                    lappend column_names $attributes(name)
                    set bind_$attributes(name) $attributes(default_value)

                    lappend column_values [dtype::sql_value \
                        $attributes(datatype) \
                        bind_$attributes(name)]

                }
            } else {

                # We are updating an existing object so the default value is
                # the old value for the row

                lappend column_names $attributes(name)
                lappend column_values $attributes(name)

            }
        }
    }

    if {[string equal $action create]} { 
        db_dml insert_statement "
        insert into ${type_db(table_name)}i (object_id, [join $column_names ", "])
        values (:object_id, [join $column_values ", "])"
    } else {
        set updates [list]

        set length [llength $column_names]
        for {set i 0} {$i < $length} {incr i} {
            lappend updates \
                "[lindex $column_names $i] = [lindex $column_values $i]"
        }

        db_dml update_statement "
        update ${type_db(table_name)}i 
        set [join $updates ", "]
        where $type_db(id_column) = :object_id"
    }

    #######################################################
    # Post-process custom elements
    #
    dtype::form::post_process -prefix $prefix \
        -ats_form $ats_form \
        -elements elements \
        -local \
        -context context

    return $object_id
}

ad_proc -callback dtype::form::process -impl content_revision {
} {
    Process a dynamic type form submission for a sub-type of content_revision,
    created by a function such as dtype::form::generate.  

    @param prefix the prefix for each element in the form
    @param ats_form the name of the template::form used
    @param form_id the form_id of the dtype form
    @param dtype_form specifies the stored object form used
    @param action whether the form is creating an object or editing one
    @param object_type the object type whose metadata will define the form
    @param object_id of the object being manipulated
    @param defaults default values to use for attributes

    @param cr_storage the storage option for the content repository
    @param cr_mime_filters the processing filter to use for specific mime types
    
    @see dtype::form::generate
} {
    set cr_storage file
    set cr_mime_Filters {text/html dtype::mime_filters::text_html}

    set argc [llength $args]
    for {set i 0} {$i <= $argc} {incr i} {
        switch -exact -- [lindex $args $i] {
            -cr_storage {
                incr i
                set cr_storage [lindex $args $i]
            }
            -cr_mime_filters {
                incr i
                set cr_mime_filters [lindex $args $i]
            }
        }
    }

    acs_object_type::get -object_type $object_type -array type_db

    array set default $defaults

    #######################################################
    # Setup default values for object creation
    #
    if {[string equal $action create]} {
        set default(object_type) $object_type

        if {![info exists default(creation_user)]} {
            set default(creation_user) [ad_conn user_id]
        }

        if {![info exists default(creation_ip)]} {
            set default(creation_ip) [ad_conn peeraddr]
        }
    }

    #######################################################
    # Setup the context for custom element callbacks
    #
    array set context [list \
        action $action \
        object_id $object_id \
        object_type $object_type \
        user_id [ad_conn user_id] \
    ]

    if {[string equal $action edit]} {
        # Get existing object data
        dtype::get_object -object_id $object_id -array object

        set context(old_values) [array get object]
    }

    #######################################################
    # Content Repository specific preparations
    #
    if {[string equal $action create]} {
        # We are creating an initial revision of a content item (ie. a new 
        # instance of a subtype of content_revision).  We need to first 
        # create a content_item object.
        set item_id [db_nextval acs_object_id_seq]

        array set item_defaults [list item_id $item_id \
                                     name "item$item_id" \
                                     locale "" \
                                     parent_id "" \
                                     content_type $object_type \
                                     creation_user [ad_conn user_id] \
                                     creation_ip [ad_conn peeraddr] \
                                     storage_type $cr_storage]
        
        foreach var [array names item_defaults] {
            if {[info exists default($var)]} {
                set item_$var $default($var)
            } else {
                set item_$var $item_defaults($var)
            }
        }

        #
        # Create the item for this for this new object
        #
        db_exec_plsql create_item {        select content_item__new(varchar :item_name,
                                 :item_parent_id,
                                 :item_item_id,
                                 :item_locale,
                                 now(),
                                 :item_creation_user,
                                 null,
                                 :item_creation_ip,
                                 'content_item',
                                 :item_content_type,
                                 null,
                                 null,
                                 'text/plain',
                                 null,
                                 null,
                                 :cr_storage)
}
    } else {
        # We are adding a revision to an existing content type - we ignore
        # any passed in storage type and use the one set in the content 
        # item
        set item_id [item::get_item_from_revision $object_id]
        item::get -item_id $item_id -array item

        set cr_storage $item(storage_type)
    }

    set context(item_id) $item_id

    # Populate content revision fields with default values
    set default(nls_language) ""
    set default(publish_date) ""

    #######################################################
    # Get the elements in the form
    #
    array set values [list]

    dtype::form::elements -local -form_id $form_id -multirow elements

    dtype::form::pre_process -prefix $prefix \
        -ats_form $ats_form \
        -elements elements \
        -local \
        -values values \
        -context context

    #######################################################
    # Perform the insert or update
    #
    template::multirow -local -key attribute_id index elements 

    # get the attribute metadata for the object type
    dtype::get_attributes -local -name $object_type \
        -start_with acs_object \
        attributes

    set column_names [list]
    set column_values [list]

    set size [template::multirow -local size attributes]
    for {set i 1} {$i <= $size} {incr i} {
        template::multirow -local get attributes $i
        
        # Get the elements row number for this attribute if it exists
        set element_row [lindex [template::multirow -local \
                    -key $attributes(attribute_id) \
                    rows \
                    elements] \
                0]

        if {![string equal $element_row ""]} {
            template::multirow -local get elements $element_row

            # This attribute is represented in the form so use the value given
            # in the form

            if {[info exists values($elements(element))]} {
                lappend column_names $attributes(name)
                set bind_$attributes(name) $values($elements(element))

                lappend column_values [dtype::sql_value \
                    $attributes(datatype) \
                    bind_$attributes(name)]
            }
        }

        if {![info exists bind_$attributes(name)]} {
            # There was no value generated by the form for this attribute so 
            # there are several ways to determine a default value
        
            if {[info exists default($attributes(name))]} {
        
                # A default was provided for this attribute so we use it
            
                lappend column_names $attributes(name)
                set bind_$attributes(name) $default($attributes(name))

                lappend column_values [dtype::sql_value \
                    $attributes(datatype) \
                    bind_$attributes(name)]

            } elseif {[string equal $action create]} {
                if {![string equal $attributes(default_value) ""]} {

                    # We are creating a new object so we use the default 
                    # value for the attribute supplied in the type metadata
                
                    lappend column_names $attributes(name)
                    lappend column_values ":bind_$attributes(name)"

                    lappend column_values [dtype::sql_value \
                        $attributes(datatype) \
                        bind_$attributes(name)]

                }
            } else {

                # We are updating an existing object so the default value is
                # the old value for the row

                lappend column_names $attributes(name)
                lappend column_values $attributes(name)

            }
        }
    }

    if {[string equal $action create]} { 
        set context(latest_revision) ""

        db_dml insert_statement "
        insert into ${type_db(table_name)}i 
        ([join [concat "item_id" "revision_id" $column_names] ", "])
        values 
        ([join [concat ":item_id" ":object_id" $column_values] ", "])"

    } else { 
        set latest_revision \
            [content::item::get_latest_revision -item_id $item_id]
        set context(latest_revision) $latest_revision

        db_dml insert_statement "
        insert into ${type_db(table_name)}i 
        ([join [concat "item_id" "revision_id" $column_names] ", "])
        select  
        [join [concat ":item_id" ":object_id" $column_values] ", "]
        from ${type_db(table_name)}i
        where revision_id = $latest_revision"

    }

    #######################################################
    # Post-process custom elements
    #
    dtype::form::post_process -prefix $prefix \
        -ats_form $ats_form \
        -elements elements \
        -local \
        -context context

    return $object_id
}

ad_proc -private dtype::sql_value {
    datatype
    variable
} {
    Returns an appropriate sql syntax value for a given datatype type.
} {
    switch -exact -- $datatype {
        date {
            return "to_date(:$variable,'YYYY-MM-DD HH24:MI:SS')"
        }
        time_of_day -
        timestamp {
            return "to_timestamp(:$variable,'YYYY-MM-DD HH24:MI:SS')"
        }
        default {
            return ":$variable"
        }
    }
}

ad_proc -private dtype::form::pre_process {
    {-prefix:required}
    {-ats_form:required}
    {-elements:required}
    {-values:required}
    {-context:required}
    {-local:boolean}
    {-ulevel 1}
} {
    Pre-process the form elements and generate values for each.
} {
    upvar $values l_values
    upvar $context l_context

    template::multirow -local=$local_p \
        -ulevel $ulevel \
        upvar \
        $elements \
        l_elements

    set size [template::multirow -local size l_elements]
    for {set e 1} {$e <= $size} {incr e} {
        template::multirow -local get l_elements $e

        if {![string equal $l_elements(custom) ""] 
            && [callback::impl_exists \
                -callback dtype::form::pre_process \
                -impl $l_elements(custom)]} {

            set l_values($l_elements(element)) [lindex \
                [callback -impl $l_elements(custom) \
                    dtype::form::pre_process \
                        -prefix $prefix \
                        -ats_form $ats_form \
                        -elements l_elements \
                        -values l_values \
                        -context l_context \
                        -local \
                        -row $e] \
                0]
        }

        if {![info exists l_values($l_elements(element))]
            && ![string equal $l_elements(widget) ""]} {
            set l_values($l_elements(element)) \
                [dtype::element::get_ats_value \
                    -prefix $prefix \
                    -ats_form $ats_form \
                    -element l_elements]
        }
    }

}

ad_proc -callback dtype::form::pre_process {
    {-prefix:required}
    {-ats_form:required}
    {-elements:required}
    {-local:boolean}
    {-ulevel 1}
    {-values:required}
    {-context:required}
    {-row:required}
} {
    Pre-processes a form element and may generate a value for it.
} -

ad_proc -private dtype::form::post_process {
    {-prefix:required}
    {-ats_form:required}
    {-elements:required}
    {-local:boolean}
    {-ulevel 1}
    {-context:required}
} {
    Post-process the form elements.
} {
    upvar $context l_context

    template::multirow \
        -local=$local_p \
        -ulevel $ulevel \
        upvar \
        $elements \
        l_elements

    set size [template::multirow -local size l_elements]
    for {set e 1} {$e <= $size} {incr e} {
        template::multirow -local get l_elements $e

        if {![string equal $l_elements(custom) ""] &&
            [callback::impl_exists \
                -callback dtype::form::post_process \
                -impl $l_elements(custom)]} {

            callback -impl $l_elements(custom) \
                dtype::form::post_process \
                    -prefix $prefix \
                    -ats_form $ats_form \
                    -elements l_elements \
                    -local \
                    -context l_context \
                    -row $e
        }
    }
}

ad_proc -callback dtype::form::post_process {
    {-prefix:required}
    {-ats_form:required}
    {-elements:required}
    {-local:boolean}
    {-ulevel 1}
    {-context:required}
    {-row:required}
} {
    Post-processes a form element.
} -

ad_proc -private dtype::element::get_ats_value {
    {-prefix:required}
    {-ats_form:required}
    {-element:required}
} {
    Get an elements value from a form.
} {
    upvar $element l_element

    switch -- $l_element(widget) {
        file {
            set value ""
        }
        checkbox -
        multiselect {
            return [template::element::get_values \
                $ats_form \
                ${prefix}$l_element(element)]
        }
        date -
        time_of_day -
        timestamp {
            return [template::util::date::get_property ansi \
                [template::element::get_value \
                    $ats_form \
                    ${prefix}$l_element(element)]]
        }
        default {
            return [template::element::get_value \
                $ats_form \
                ${prefix}$l_element(element)]
        }
    }
}

ad_proc -private dtype::form::value_switch {
    {-widget:required}
    {-value:required}
} {
    Return a -value or -values switch appropriately 
} {
    # always a space before -value as it's appended directly to the switches
    switch $widget {
        file {}
        checkbox -
        multiselect {
            return " -values $value"
        }
        default {
            return " -value $value"
        }
    }

    return ""
}

ad_proc -private dtype::upload_content {
    {-item_id:required}
    {-revision_id:required}
    {-file:required}
    {-storage_type:required}
} {
    Upload the content in file for the specified revision and store using
    the method specified.
} {
    if {[string equal $storage_type file]} {
        set file_path [cr_create_content_file $item_id $revision_id $file]
        set file_size [file size $file]
        db_dml upload_file_revision {}
    } elseif {[string equal $storage_type text]} {
        # upload the file into the revision content
        db_dml upload_text_revision {} -blob_files [list $file]
    } else {
        # upload the file into the revision content
        db_dml upload_revision {} -blob_files [list $file]
    }
}

ad_proc -private dtype::write_utf8_file {
    content
} {
    Write a temporary file in utf-8 character encoding containing the text 
    supplied.
} {
    set tmp_file [ns_tmpnam]

    set fd [open $tmp_file w]
    fconfigure $fd -encoding utf-8

    puts $fd $content
    
    close $fd
    return $tmp_file
}

# Templates 
#
# TODO: Widget templates are currently out of date and therefore don't work at
# TODO  all.  They need to be updated to be something sane.
#

ad_proc -public dtype::form::widget_templates {
    {-datatypes {}}
    multirow
} {
    set datatype_clause ""

    if {[llength $datatypes] > 0} {
        set datatype_clause "and wt.datatype in ('[join $datatypes "', '"]')"
    }
    
    db_multirow $multirow select_widget_templates {}
}

ad_proc -public dtype::form::widget_template {
    {-template:required}
    array
} {
    Get a widget template.
} {
    upvar $array result

    db_1row select_widget_template {} -column_array result
}

ad_proc -public dtype::form::widget_template_params {
    {-template:required}
    multirow
} {
    db_multirow -local -upvar_level 2 $multirow select_widget_template_params {}
}

ad_proc -public dtype::form::create_widget_param {
    {-object_type:required}
    {-dform:required}
    {-attribute_name:required}
    {-param_name:required}
    {-type:required}
    {-source:required}
    {-value:required}
} {
    Create a parameter for the specified widget.
} {
    db_exec_plsql create_widget_param {}
}

ad_proc -public dtype::form::clone_widget_template {
    {-object_type:required}
    {-dforms:required}
    {-attribute_name:required}
    {-template_name:required}
    {-required_p 0}
} {
    Create a widget and corresponding parameters according to the
    specified template.
} {
    dtype::form::widget_template -template $template_name template 
    dtype::form::widget_template_params -template $template_name \
        template_params

    set size [template::multirow size template_params]

    foreach dform $dforms {
        dtype::form::create_widget \
            -object_type $object_type \
            -dform $dform \
            -attribute_name $attribute_name \
            -widget $template(widget) \
            -required_p $required_p 

        for {set i 1} {$i <= $size} {incr i} {
            template::multirow get template_params $i

            dtype::form::create_widget_param \
                -object_type $object_type \
                -dform $dform \
                -attribute_name $attribute_name \
                -param_name $template_params(param) \
                -type $template_params(param_type) \
                -source $template_params(param_source) \
                -value $template_params(value)
        }        
    } 
} 

ad_proc -public dtype::form::delete_attribute_widgets {
    {-object_type:required}
    {-attribute_name:required}
} {
    Delete widgets associated with a specified attribute.
} {
    db_foreach get_widget_forms {} {
        dtype::form::delete_widget -object_type $object_type \
            -attribute_name $attribute_name \
            -dform $dform
    }
}

#
# Declarative API
#

ad_proc -public dtype::spec::create_form { spec } {
    Create a form from a specification
} {
  foreach { form_type form_spec_list } $spec {
    array set form_spec $form_spec_list

    set cmd "dtype::form::create"

    # Append optional arguments
    foreach name {
        layout
        widget
        custom
    } {
        if {[info exists form_spec($name)]} {
            append cmd " -$name \"$form_spec($name)\""
        }
    }

    set form_id [eval $cmd]

    dtype::form::map_type -form_id $form_id \
        -form $form_type \
        -object_type $form_spec(object_type)

    dtype::get_attributes -local -name $form_spec(object_type) \
        -start_with acs_object \
        attributes

    dtype::spec::create_element \
        -attributes attributes \
        -local \
        -parent_id [dtype::form::get_root_element_id -form_id $form_id] \
        $form_spec(elements)
  }
}

ad_proc -private dtype::spec::create_element { 
    {-parent_id:required}
    {-attributes:required}
    {-local:boolean}
    {-ulevel 1}
    spec 
} {
    Adds an element to a form created from a spec.
} {
    template::multirow -local=$local_p \
        -ulevel $ulevel \
        upvar \
        $attributes \
        l_attributes

    template::multirow -local \
        -key name \
        index \
        l_attributes

    foreach { element_type element_spec_list } $spec {
      array set element_spec $element_spec_list

      set cmd "dtype::element::add_element -element \"$element_type\" \
        -parent_id \"$parent_id\""

      # Append optional arguments
      foreach name {
          layout
          widget
          custom
      } {
          if {[info exists element_spec($name)]} {
              append cmd " -$name \"$element_spec($name)\""
          }
      }

      set element_id [eval $cmd]
      
      dtype::spec::create_parameter \
        -proc template::element::create \
        -call_id $element_id \
        $element_spec(params)

      if {[info exists element_spec(attribute)]} {
          template::multirow -local \
            -key $element_spec(attribute) \
            get \
            l_attributes

          dtype::element::map_attribute -element_id $element_id \
            -attribute_id $l_attributes(attribute_id)
      }

      if {[info exists element_spec(elements)]} {
          dtype::spec::create_element \
            -attributes l_attributes \
            -local \
            -parent_id $element_id \
            $element_spec(elements)
      }

      unset element_spec
    }
}

