ad_library {
    A library of functions to create and manipulate dtype procs.

    @author Lee Denison (lee@xarg.net)
    @creation-date 2005/12/06
} 

namespace eval dtype {}
namespace eval dtype::proc {}
namespace eval dtype::spec {}

ad_proc -public dtype::proc::add_parameter_value {
    {-proc:required}
    {-param:required}
    {-call_id:required}
    {-value:required}
    {-source literal}
} {
    Adds a parameter value to a call to a proc.
} {
    db_exec_plsql add_parameter {}
}

ad_proc -public dtype::proc::params {
    {-proc:required}
    {-local:boolean}
    {-ulevel 1}
    {-multirow:required}
} {
    Creates a multirow of possible parameters for a proc.
} {
    set keys [list \
      param_id \
      name \
      param_type \
      is_required \
      default_value \
    ]

    set multirow_cmd "template::multirow"

    if {$local_p} {
        append multirow_cmd " -local -ulevel [expr {$ulevel + 1}]"
    }

    eval "$multirow_cmd -key name create \$multirow $keys"

    set metadata [dtype::proc::params_list \
        -proc $proc]
    
    foreach param $metadata {
        eval "$multirow_cmd append \$multirow $param"
    }
}

ad_proc -private dtype::proc::params_list {
    {-proc:required}
} {
    Returns a list of lists with the proc parameter metadata for the 
    specified proc.
} {
    return [db_list_of_lists select_proc_metadata {}]
}

ad_proc -public dtype::proc::param_values {
    {-proc:required}
    {-call_id:required}
    {-local:boolean}
    {-ulevel 1}
    {-multirow:required}
} {
    Creates a multirow of parameter values for a call indexed by name.
} {
    set keys [list \
      value_id \
      call_id \
      param_id \
      name \
      param_type \
      param_source \
      is_required \
      default_value \
      value \
    ]

    set multirow_cmd "template::multirow"

    if {$local_p} {
        append multirow_cmd " -local -ulevel [expr {$ulevel + 1}]"
    }

    eval "$multirow_cmd -key name create \$multirow $keys"

    set metadata [dtype::proc::param_values_list \
        -proc $proc \
        -call_id $call_id]
    
    foreach param $metadata {
        eval "$multirow_cmd append \$multirow $param"
    }
}

ad_proc -private dtype::proc::param_values_list {
    {-proc:required}
    {-call_id:required}
} {
    Returns a list of lists with the proc parameter metadata for the 
    specified call.
} {
    return [db_list_of_lists select_proc_metadata {}]
}

ad_proc -public dtype::proc::call {
    {-proc:required}
    {-call_id:required}
    {-context {}}
} {
    Make a call to a stored proc with the specified call parameters.
} {
    if {[string equal $context ""]} {
        array set l_context [list]
    } else {
        upvar $context l_context
    }

    dtype::proc::params -proc $proc \
        -multirow params \
        -local 

    dtype::proc::param_values -proc $proc \
        -call_id $call_id \
        -multirow param_values \
        -local

    # Create an array for spec switches, which are built up from several
    # parameters
    array set spec_switches [list]
    set switches [list]
    set bool_switches [list]

    # Generate the parameters
    set size [template::multirow -local size params]
    for {set p 1} {$p <= $size} {incr p} {
        template::multirow -local get params $p

        set value_rows [template::multirow -local \
            -key $params(name) \
            rows \
            param_values]

        if {[llength $value_rows] > 0} {

            # Use the parameter value stored in the database
            template::multirow -local -key $params(name) get param_values 

            set value [dtype::proc::param_value \
                -context l_context \
                -parameter params \
                -value param_values]

        } elseif {[info exists $params(default_value)]} {

            # Use the default value stored in the database
            set value $params(default_value)

        }

        if {[info exists value]} {
            switch -- $params(param_type) {
                switch:onevalue {
                    lappend switches "-$params(name)" $value
                }
                switch:boolean {
                    if {$value} {
                        lappend bool_switches "-$params(name)"
                    }
                }
                switch:spec {
                    set name_list [split $params(name) ","]
                    set major [lindex $name_list 0]
                    set minor [lindex $name_list 1]

                    lappend spec_switches($major) $minor $value
                }
            }
        }
    }
        
    foreach name [array names spec_switches] {
        append proc " -$name {$spec_switches($name)}"
    }
     
    return [eval "$proc $bool_switches $switches"]
}

ad_proc -public dtype::proc::param_value {
    {-context:required}
    {-parameter:required}
    {-value:required}
} {
    Calculates and returns the current value for the supplied parameter array 
    based on its source and default_value attributes.
} {
    upvar $context __context
    upvar $parameter __param
    upvar $value __value

    foreach __name [array names __context] {
        set $__name $__context($__name)
    }

    set __result ""

    switch $__value(param_source) {
        scope {

            # A value passed in context
            set __result [set $__value(value)]

        }
        proc {
            if [catch {

                # Make a nested call to determine the parameter value
                set __result [dtype::proc::call -proc $__value(value) \
                    -call_id $__value(value_id) \
                    -context __context]

            } error] {
                global errorInfo

                set name $__param(param)
                ns_log warning "[_ dynamic-types.unable_to_retrive_param]: $error"
                ns_log debug "$errorInfo"

                set __result ""
            }
        }
        literal -
        default {

            # A literal value stored in the database
            set __result $__value(value)

        }
    }

    return $__result
}

ad_proc -private dtype::spec::create_parameter {
    {-proc:required}
    {-call_id:required}
    spec
} {
    Create a parameter from a spec.
} {
    foreach { param_type param_spec_list } $spec {
        array set param_spec $param_spec_list

        set cmd "dtype::proc::add_parameter_value -proc \"$proc\" \
            -call_id \"$call_id\" \
            -param \"$param_type\" \
            -value \"$param_spec(value)\""

        # Append optional arguments
        foreach name {
            source
        } {
            if {[info exists param_spec($name)]} {
                append cmd " -$name \"$param_spec($name)\""
            }
        }

        set param_id [eval $cmd]

        if {[info exists param_spec(params)]} {
            dtype::spec::create_parameter \
              -proc $param_spec(value) \
              -call_id $param_id \
              $param_spec(params)
        }
    }
}
